/*
 Copyright:      Radig Ulrich  mailto: mail@ulrichradig.de
 Author:         Radig Ulrich
 Remarks:        
 known Problems: none
 Version:        28.05.2004
 Description:    Ansteuerung eines Block Device via FAT
*/

#include <fat.h>

//Prototypes
long Root_Dir_Addr (void);
int Read_Root_Dir_Ent (int,char *Buffer);
long FAT_Addr (void);
void Load_File (int Cluster,char *Buffer);
void Read_File (char *FatTable,char *Buffer,long ,char,int);
char unsigned ClusSiz (void);
int unsigned ClusOff (void);

//Auslesen der Adresse des Volume Boot Record von MBR
long FAT_Addr (void)
{
	char Buffer[BlockSize];
	unsigned long Volume_Boot_Record_Addr;

	//auslesen des Master Boot Record von der MMC/SD Karte (addr = 0)
	Read_Block_MMC (MasterBootRecord,Buffer); //Read Master Boot Record

	//Berechnet Volume Boot Record 
	Volume_Boot_Record_Addr  = (Buffer[VBR_Addr+1] << 8) + (Buffer[VBR_Addr]);
	return (Volume_Boot_Record_Addr);
}

//Auslesen der Adresse des First Root Directory von Volume Boot Record
long Root_Dir_Addr (void)
{
	char Buffer[BlockSize];
	unsigned long Volume_Boot_Record_Addr = FAT_Addr ();
	unsigned long FirstRootDirSecNum;
	
	//auslesen des Volume Boot Record von der MMC/SD Karte (addr = Volume_Boot_Record_Addr)
	Read_Block_MMC (Volume_Boot_Record_Addr,Buffer);
	
	//berechnet den ersten Sector des Root Directory
	FirstRootDirSecNum = ((Buffer[BPB_RsvdSecCnt+1] << 8)+Buffer[BPB_RsvdSecCnt]) + (Buffer[BPB_NumFATs] * ((Buffer[BPB_FATSz16+1] << 8)+Buffer[BPB_FATSz16]));
	FirstRootDirSecNum+= Volume_Boot_Record_Addr;
	return(FirstRootDirSecNum);
}

/*	Ausgabe des angegebenen Directory Eintrag in Entry_Count
	ist kein Eintrag vorhanden, ist der Eintrag im 
	Rckgabe Cluster 0xFFFF. Es wird immer nur ein Eintrag ausgegeben
	um Speicherplatz zu sparen um es auch fr kleine Atmels zu benutzen
*/

int Read_Root_Dir_Ent (int Entry_Count,char *Buffer)
{
	char TMP_Buffer[BlockSize];
	int TMP_Entry_Count = 0;
	char Cluster = 0;
	//Auslesen der Bolcklnge des Root Directory
	unsigned long Volume_Boot_Record_Addr = FAT_Addr ();
	Read_Block_MMC (Volume_Boot_Record_Addr,TMP_Buffer);	
	unsigned long RootDirSectors = ((TMP_Buffer[BPB_RootEntCnt + 1] << 8) + TMP_Buffer[BPB_RootEntCnt])*32;
	RootDirSectors = RootDirSectors + (((TMP_Buffer[BPB_BytesPerSec + 1] << 8) + TMP_Buffer[BPB_BytesPerSec])-1);
	RootDirSectors = RootDirSectors / ((TMP_Buffer[BPB_BytesPerSec + 1] << 8) + TMP_Buffer[BPB_BytesPerSec]);

	//Auslesen des Blocks wo das Root Directory beginnt
	long FirstRootDirSecNum = Root_Dir_Addr();

	//auslesen des gesamten Root Directory
	for (long b = FirstRootDirSecNum;b<(FirstRootDirSecNum+RootDirSectors);b++)
	{
		Read_Block_MMC (b,TMP_Buffer);	//Lesen eines Blocks des Root Directory
	
		for (int a=0;a<BlockSize;a=a+32)
		{
			if ((TMP_Buffer[a] != 0x42) & (TMP_Buffer[a] > 0x20) & (TMP_Buffer[a] != 0xE5)) //Prfen ob es ein 8.3 Eintrag ist
			{
				TMP_Entry_Count++;
				if (TMP_Entry_Count == Entry_Count) //Ist es der gewnschte Verzeichniseintrag
				{
					//Speichern des Verzeichnis Eintrages in den Rckgabe Buffer
					for(char b=DIR_Name;b<11;b++)
					{
					*Buffer++=TMP_Buffer[a+b];
					}						
					*Buffer++='\0';
					
					*Buffer++=TMP_Buffer[a+DIR_Attr];
					
					*Buffer++=TMP_Buffer[a+DIR_FileSize+3];
					*Buffer++=TMP_Buffer[a+DIR_FileSize+2];
					*Buffer++=TMP_Buffer[a+DIR_FileSize+1];
					*Buffer++=TMP_Buffer[a+DIR_FileSize];
					//Speichern des Clusters des Verzeichniseintrages
					Cluster = ((TMP_Buffer[a+DIR_FstClusLO+1] << 8) + TMP_Buffer[a+DIR_FstClusLO]);
					return(Cluster);//Eintrag gefunden Rcksprung mit Cluster File Start
				}
			}
		}
	}
	return (0xFFFF); //Kein Eintrag mehr gefunden Rcksprung mit 0xFFFF
}

/*	Auslesen der Cluster fr ein File aus der FAT
	in den Buffer(512Byte). Bei einer 128MB MMC/SD 
	Karte ist die Cluster gre normalerweise 16KB gro
	das bedeutet das File kann max. 4MByte gro sein.
	Bei greren Files mu der Buffer grer definiert
	werden! (Ready)
*/	
void Load_File (int Cluster,char *Buffer) //Cluster = Start Clusterangabe aus dem Directory
{
	unsigned int FAT_Block_Store = 0;	//Zum berprfen ob der FAT Bolck schon geladen wurde
	unsigned int FAT_Byte_Addresse;	//Byte Adresse innerhalb des Fat Blocks
	unsigned int FAT_Block_Addresse;	//FAT Block Adresse
	char TMP_Buffer[BlockSize];
	
	//Berechnung fr den ersten FAT Bolck (FAT Start Addresse)
	unsigned int Volume_Block_Start = FAT_Addr();
	Read_Block_MMC (Volume_Block_Start,TMP_Buffer);		
	unsigned int FAT_Offset = (TMP_Buffer[BPB_RsvdSecCnt + 1] << 8) + TMP_Buffer[BPB_RsvdSecCnt];

	for (int a = 0;a<(BlockSize/2);a++)
	{
		//Speichern des Clusters in den Buffer
		*Buffer++ = (Cluster & 0xFF00)>>8; //MSB
		*Buffer++ = (Cluster & 0x00FF);	//LSB
		
		if (Cluster == 0xFFFF)
		{
		break; //Ist das Ende des Files erreicht Schleife beenden
		}
		//Berechnung des Bytes innerhalb des FAT Blocks
		FAT_Byte_Addresse = (Cluster*2) % BlockSize;

		//Berechnung des Blocks der gelesen werden mu
		FAT_Block_Addresse = ((Cluster*2) / BlockSize) + Volume_Block_Start + FAT_Offset;	

		//Lesen des FAT Blocks
		//berprfung ob dieser Block schon gelesen wurde
		if (FAT_Block_Addresse != FAT_Block_Store)
			{
			FAT_Block_Store = FAT_Block_Addresse;
			//Lesen des FAT Blocks
			Read_Block_MMC (FAT_Block_Addresse,TMP_Buffer);	
			}
		//Lesen des Clusters
		Cluster = (TMP_Buffer[FAT_Byte_Addresse + 1] << 8) + TMP_Buffer[FAT_Byte_Addresse];		
	}
}

//Read Cluster Size
char unsigned ClusSiz (void)
{
	char Buffer[BlockSize];
	int ClusterSize;
	unsigned long Volume_Boot_Record_Addr = FAT_Addr ();	
	//auslesen des Volume Boot Record von der MMC/SD Karte (addr = Volume_Boot_Record_Addr)
	Read_Block_MMC (Volume_Boot_Record_Addr,Buffer);
	ClusterSize = Buffer[BPB_SecPerClus];
	return (ClusterSize);
}

int unsigned ClusOff (void)
{
	char Buffer[BlockSize];
	int ClusterOffset = Root_Dir_Addr();
	unsigned long Volume_Boot_Record_Addr = FAT_Addr ();	
	//auslesen des Volume Boot Record von der MMC/SD Karte (addr = Volume_Boot_Record_Addr)
	Read_Block_MMC (Volume_Boot_Record_Addr,Buffer);
	ClusterOffset += (((Buffer[BPB_BytesPerSec + 1] << 8) + Buffer[BPB_BytesPerSec]) * 32/BlockSize);
	return (ClusterOffset);
}

void Read_File (char *FATTabel,char *Buffer,long BlockCount,char ClusterSize,int ClusterOffset)
{
	//Berechnung des Blocks aus BlockCount und Cluster aus FATTabelle
	//Berechnung welcher Cluster zu laden ist
	unsigned long Block = (FATTabel[((BlockCount/ClusterSize)*2)]<<8)+FATTabel[((BlockCount / ClusterSize)*2)+1]-2;
	Block = (Block * ClusterSize) + ClusterOffset;
	//Berechnung des Blocks innerhalb des Cluster
	Block += (BlockCount % ClusterSize);
	//Read Data Block from Device
	Read_Block_MMC (Block,Buffer);	
}


